########## R function: AsyStDevJWB ##########

# For estimating the asymptotic variances of the
# components of the "betaB" parameter vector for
# logistic and Poisson generalized linear mixed models,
# based on the asymptotic normality theorem of
# Jiang, Wand & Bhaskaran (2020). This version is for
# the case of a random intercept.

# Last changed: 10 SEP 2021

AsyStDevJWB <- function(beta0Hat,betaBhat,sigsqHat,XB,m,family,
                         numBisect=20,suppTolerance=1e-15,ngTpz=1001)
{
   # Check legality of the family argument:
 
   if (!any(family==c("binomial","poisson")))
      stop("The family must be one of Bernoulli or Poisson.\n")   

   # Define a function for trapezoidal integration:

   trapint <- function(xgrid,fgrid)
   {
      ng <- length(xgrid)
      xvec <- xgrid[2:ng] - xgrid[1:(ng-1)]
      fvec <- fgrid[1:(ng-1)] + fgrid[2:ng]
      return(sum(xvec*fvec)/2)
   }

   if (family=="poisson")
   {
      # Obtain required (X^T)W(X)-type matrix:

      wtVec <- exp(beta0Hat + 0.5*sigsqHat + as.vector(XB%*%betaBhat))
      XBwIncpt <- cbind(1,XB)
      XTWX <- crossprod(XBwIncpt*wtVec,XBwIncpt)
    
      # Obtain the asymptotic covariance matrix of betaBhat:

      dB <- ncol(XB)
      AsyCOVbetaBhat <- solve(XTWX)[(2:(dB+1)),(2:(dB+1))]

      # Return the vector of asymptotic standard deviations:

      return(sqrt(c((sigsqHat/m),diag(AsyCOVbetaBhat))))
   }
  
   if (family=="binomial")
   {
      # Set the bddLogis() and bdddLogis() functions:

      bddLogis <- function(x) return(0.5/(1 + cosh(x)))
      bdddLogis <- function(x) return(-0.5*sinh(x)/((1 + cosh(x))^2))

      # Set the (k,kd) entry integrand function and its derivative:

      intgd <- function(u,beta0Hat,betaBhat,sigsqHat,XB,XbetaBhat,k,kd)
      {
         dB <- ncol(XB)
         XBwIncpt <- cbind(1,XB)
         LRinds <- (2:(dB+1))
         wVec <- bddLogis((beta0Hat+u) + XbetaBhat)
         Theta0 <- crossprod((XBwIncpt*wVec),XBwIncpt)   

         return(exp(-0.5*u*u/sigsqHat)*solve(solve(Theta0)[LRinds,LRinds])[k,kd])
      } 

      intgdDrv <- function(u,beta0Hat,betaBhat,sigsqHat,XB,XbetaBhat,k,kd)
      {
         dB <- ncol(XB)
         XBwIncpt <- cbind(1,XB)
         LRinds <- (2:(dB+1))
         wVec <- bddLogis((beta0Hat+u)*rep(1,n)+XbetaBhat)
         wdVec <- bdddLogis((beta0Hat+u)+XbetaBhat)
         Theta0 <- crossprod((XBwIncpt*wVec),XBwIncpt) 
         Thetad0 <- crossprod((XBwIncpt*wdVec),XBwIncpt) 
         Theta1 <- solve(Theta0)[LRinds,LRinds]
         Theta2 <- solve(Theta0,(Thetad0%*%solve(Theta0)))
         Theta2 <- Theta2[LRinds,LRinds]

         return(exp(-0.5*u*u/sigsqHat)*solve(Theta1,t(solve(t(Theta1),Theta2)) 
                - (u/sigsqHat)*diag(dB))[k,kd])
      } 

      # Extract the dimension variables:

      n <- nrow(XB)   ;   dB <- ncol(XB)

      # Fill up the entries of invLambdaHatBetaB:

      invLambdaHatBetaB <- matrix(NA,dB,dB)
      XbetaBhat <- as.vector(crossprod(t(XB),betaBhat))
      for (k in 1:dB)
      {
         for (kd in k:dB) 
         {
            # Obtain the root of the derivative of the log-integrand function using 
            # bisection searching:

            needBisecSearch <- TRUE
            rootCaptured <- FALSE

            uLow <- -1 - 1/pi  ;   uUpp <- 1 + exp(-1)
            while (!rootCaptured)
            {
               hdLow <- intgdDrv(uLow,beta0Hat,betaBhat,sigsqHat,XB,XbetaBhat,k,kd)
               hdUpp <- intgdDrv(uUpp,beta0Hat,betaBhat,sigsqHat,XB,XbetaBhat,k,kd)
   
               # Take care of cases where root corresponds
               # to hdLow or hdUpp:
   
               if (hdLow==0)
               {
                  rootCaptured <- TRUE
                  needBisecSearch <- FALSE
                  uMid <- uLow
               }

               if (hdUpp==0)
               {  
                  rootCaptured <- TRUE
                  needBisecSearch <- FALSE
                  uMid <- uUpp
               }
         
               # Now deal with the more usual case
               # where bisection searching is needed:

               if (hdLow<0) uLow <- 2*uLow
               if (hdUpp>0) uUpp <- 2*uUpp

               if ((hdLow>0)&(hdUpp<0)) rootCaptured <- TRUE
            }   

            if (needBisecSearch)
            {
               for (ib in 1:numBisect)
               {
                  uMid <- (uLow + uUpp)/2
                  hdMid <- intgdDrv(uMid,beta0Hat,betaBhat,sigsqHat,XB,XbetaBhat,k,kd)
                  if (hdMid>0) uLow <- uMid
                  if (hdMid<0) uUpp <- uMid
               }
               uMode <- uMid
            }

            # Obtain the effective support of the integrand:

            suppLow <- uMode - 5*sqrt(sigsqHat)
            suppUpp <- uMode + 5*sqrt(sigsqHat)
            suppChosen <- FALSE
            while(!suppChosen)
            {
               tailRatioLow <- abs((intgd(suppLow,beta0Hat,betaBhat,sigsqHat,XB,XbetaBhat,k,kd)
                               /intgd(uMode,beta0Hat,betaBhat,sigsqHat,XB,XbetaBhat,k,kd)))

               tailRatioUpp <- abs((intgd(suppUpp,beta0Hat,betaBhat,sigsqHat,XB,XbetaBhat,k,kd)
                                  /intgd(uMode,beta0Hat,betaBhat,sigsqHat,XB,XbetaBhat,k,kd)))

               tailRatio <- max(tailRatioLow,tailRatioUpp)
               
               if (tailRatio<suppTolerance) suppChosen <- TRUE
               if (tailRatio>=suppTolerance) 
               {
                  suppLow <- suppLow - sqrt(sigsqHat)
                  suppUpp <- suppUpp + sqrt(sigsqHat)
               }  
            }

            ug <- seq(suppLow,suppUpp,length=ngTpz)
            fug <- rep(NA,ngTpz)
            for (ig in 1:ngTpz)
               fug[ig] <- intgd(ug[ig],beta0Hat,betaBhat,sigsqHat,XB,XbetaBhat,k,kd)

            invLambdaHatBetaB[k,kd] <- trapint(ug,fug)
            invLambdaHatBetaB[kd,k] <- invLambdaHatBetaB[k,kd]
         }
      }
      
      # Obtain the asymptotic covariance matrix of betaBhat:

      AsyCOVbetaBhat <- sqrt(2*pi*sigsqHat)*solve(invLambdaHatBetaB)
     
      # Return the vector of asymptotic standard deviations:

      return(sqrt(c((sigsqHat/m),diag(AsyCOVbetaBhat))))
   }
}

############ End of AsyStDevJWB  ###########


 