########## R script: JWBinferenceExamples ##########

# For conducting several examples that illustrate
# inference for generalized linear mixed model 
# parameters based on the asymptotic normality
# theorem of Jiang, Wand & Bhaskaran (2020).

# Last changed: 26 OCT 2020

# Load required package and functions:

library(lme4)
source("generExampData.r")
source("AsyStDevJWB.r")
source("graphicCompareCIs.r")

# Set the example number:

exampleNumber <- 4

# Set values corresponding to the specified
# example:

if (exampleNumber==1)
{
   family <- "poisson"
   m <- 100 ; nVec <- rep(10,m)
   beta0True <- 1.2
   betaBTrue <- 0.4
   sigsqTrue <- 0.25
   Xdistn <- "unif"
}
if (exampleNumber==2)
{
   family <- "poisson"
   m <- 100 ; nVec <- rep(10,m)
   beta0True <- -0.7
   betaBTrue <- c(0.4,1.1)
   sigsqTrue <- 0.1
   Xdistn <- "unif"
}
if (exampleNumber==3)
{
   family <- "binomial"
   m <- 100 ; nVec <- rep(10,m)
   beta0True <- -1.3
   betaBTrue <- 0.1
   sigsqTrue <- 0.8
   Xdistn <- "unif"
}
if (exampleNumber==4)
{
   family <- "binomial"
   m <- 100 ; nVec <- rep(10,m)
   beta0True <- 0.02
   betaBTrue <- c(1.3,-0.4)
   sigsqTrue <- 1
   Xdistn <- "unif"
}

# Obtain data:

if (any(exampleNumber==(1:4)))
   allData <- generExampData(m,nVec,Xdistn,beta0True,betaBTrue,sigsqTrue,family)

if (exampleNumber==5)
{
   # This example involves actual data concerning
   # immunisations received by children in Guatemala.

   family <- "binomial" 
   library(mlmRev)
   data(guImmun)
   y <- as.numeric(as.character(guImmun$immun)=="Y")
   x1 <- as.numeric(as.character(guImmun$kid2p)=="Y")
   x2 <- as.numeric(as.character(guImmun$momEd)=="S")
   x3 <- as.numeric(as.character(guImmun$husEd)=="S")
   x4 <- as.numeric(as.character(guImmun$momWork)=="Y")
   x5 <- as.numeric(as.character(guImmun$rural)=="Y")
   x6 <- guImmun$pcInd81

   XB <- cbind(x1,x2,x3,x4,x5,x6)

   idNumOrig <- guImmun$mom
   idNum <- match(idNumOrig,unique(idNumOrig))

   allData <- list(y=y,XB=XB,idNum=idNum)

   m <- length(unique(allData$idNum))
}

if (!any(exampleNumber==(1:5)))
   stop("The value of `exampleNumber' must be either 1,2,3,4 or 5.\n")

# Obtain the maximimum likelihood fit using the lme4:::glmer() function:

fitMLE <- glmer(y ~ XB + (1|idNum),data=allData,nAGQ=100,family=family)

# Extract the fixed effects estimates and standard deviations 
# based on exact maximum likelihood:

betaMLEtable <- coefficients(summary(fitMLE))
betaMLE <- as.numeric(betaMLEtable[,1])
SDbetaXCT <- as.numeric(betaMLEtable[,2])
sigsqMLE <- as.numeric(attr(summary(fitMLE)$varcor$idNum,"stddev"))^2

# Obtain the approximate standard deviations based on the
# Jiang, Wand & Bhaskaran asymptotic normality theorem:

beta0MLE <- betaMLE[1] ; betaBMLE <- betaMLE[-1]
SDbetaJWB <- AsyStDevJWB(beta0MLE,betaBMLE,sigsqMLE,allData$XB,m,family=family)

# Graphically compare confidence intervals from the two approaches:

graphicCompareCIs(betaMLE,SDbetaXCT,SDbetaJWB,beta0True,betaBTrue)

############ End of JWBinferenceExamples ##########